//------------------------------------------------------------------------------
// LG_BreadthFirstSearch_vanilla_Extended:  BFS using only GraphBLAS API
//------------------------------------------------------------------------------

// LAGraph, (c) 2019-2025 by The LAGraph Contributors, All Rights Reserved.
// SPDX-License-Identifier: BSD-2-Clause
//
// For additional details (including references to third party source code and
// other files) see the LICENSE file or contact permission@sei.cmu.edu. See
// Contributors.txt for a full list of contributors. Created, in part, with
// funding and support from the U.S. Government (see Acknowledgments.txt file).
// DM22-0790

// Contributed by Scott McMillan, derived from examples in the appendix of
// The GraphBLAS C API Specification, v1.3.0

//------------------------------------------------------------------------------

// This is a Basic algorithm (no extra cached properties are required),
// but it is not user-callable (see LAGr_BreadthFirstSearch instead).

#define LG_BFS_EXTENDED
#include "template/LG_BreadthFirstSearch_vanilla_template.c"

