//------------------------------------------------------------------------------
// GB_aop__include.h: definitions for GB_aop__*.c
//------------------------------------------------------------------------------

// SuiteSparse:GraphBLAS, Timothy A. Davis, (c) 2017-2025, All Rights Reserved.
// SPDX-License-Identifier: Apache-2.0

// This file has been automatically generated from Generator/GB_aop.h
#include "math/GB_math.h"

GrB_Info GB (_subassign_23__second_bool)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__second_bool)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__second_int8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__second_int8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__second_int16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__second_int16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__second_int32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__second_int32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__second_int64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__second_int64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__second_uint8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__second_uint8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__second_uint16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__second_uint16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__second_uint32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__second_uint32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__second_uint64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__second_uint64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__second_fp32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__second_fp32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__second_fp64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__second_fp64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__second_fc32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__second_fc32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__second_fc64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__second_fc64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__pair_bool)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__pair_bool)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__pair_int8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__pair_int8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__pair_int16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__pair_int16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__pair_int32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__pair_int32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__pair_int64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__pair_int64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__pair_uint8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__pair_uint8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__pair_uint16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__pair_uint16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__pair_uint32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__pair_uint32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__pair_uint64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__pair_uint64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__pair_fp32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__pair_fp32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__pair_fp64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__pair_fp64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__pair_fc32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__pair_fc32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__pair_fc64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__pair_fc64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__min_int8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__min_int8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__min_int16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__min_int16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__min_int32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__min_int32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__min_int64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__min_int64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__min_uint8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__min_uint8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__min_uint16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__min_uint16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__min_uint32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__min_uint32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__min_uint64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__min_uint64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__min_fp32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__min_fp32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__min_fp64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__min_fp64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__max_int8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__max_int8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__max_int16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__max_int16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__max_int32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__max_int32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__max_int64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__max_int64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__max_uint8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__max_uint8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__max_uint16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__max_uint16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__max_uint32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__max_uint32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__max_uint64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__max_uint64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__max_fp32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__max_fp32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__max_fp64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__max_fp64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__plus_int8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__plus_int8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__plus_int16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__plus_int16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__plus_int32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__plus_int32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__plus_int64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__plus_int64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__plus_uint8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__plus_uint8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__plus_uint16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__plus_uint16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__plus_uint32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__plus_uint32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__plus_uint64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__plus_uint64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__plus_fp32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__plus_fp32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__plus_fp64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__plus_fp64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__plus_fc32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__plus_fc32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__plus_fc64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__plus_fc64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__minus_int8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__minus_int8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__minus_int16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__minus_int16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__minus_int32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__minus_int32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__minus_int64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__minus_int64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__minus_uint8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__minus_uint8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__minus_uint16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__minus_uint16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__minus_uint32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__minus_uint32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__minus_uint64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__minus_uint64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__minus_fp32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__minus_fp32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__minus_fp64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__minus_fp64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__minus_fc32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__minus_fc32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__minus_fc64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__minus_fc64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__rminus_int8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__rminus_int8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__rminus_int16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__rminus_int16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__rminus_int32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__rminus_int32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__rminus_int64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__rminus_int64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__rminus_uint8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__rminus_uint8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__rminus_uint16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__rminus_uint16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__rminus_uint32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__rminus_uint32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__rminus_uint64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__rminus_uint64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__rminus_fp32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__rminus_fp32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__rminus_fp64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__rminus_fp64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__rminus_fc32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__rminus_fc32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__rminus_fc64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__rminus_fc64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__times_int8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__times_int8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__times_int16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__times_int16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__times_int32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__times_int32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__times_int64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__times_int64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__times_uint8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__times_uint8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__times_uint16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__times_uint16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__times_uint32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__times_uint32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__times_uint64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__times_uint64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__times_fp32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__times_fp32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__times_fp64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__times_fp64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__times_fc32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__times_fc32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__times_fc64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__times_fc64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__div_int8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__div_int8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__div_int16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__div_int16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__div_int32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__div_int32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__div_int64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__div_int64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__div_uint8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__div_uint8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__div_uint16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__div_uint16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__div_uint32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__div_uint32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__div_uint64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__div_uint64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__div_fp32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__div_fp32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__div_fp64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__div_fp64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__div_fc32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__div_fc32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__div_fc64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__div_fc64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__rdiv_int8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__rdiv_int8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__rdiv_int16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__rdiv_int16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__rdiv_int32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__rdiv_int32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__rdiv_int64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__rdiv_int64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__rdiv_uint8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__rdiv_uint8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__rdiv_uint16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__rdiv_uint16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__rdiv_uint32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__rdiv_uint32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__rdiv_uint64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__rdiv_uint64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__rdiv_fp32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__rdiv_fp32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__rdiv_fp64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__rdiv_fp64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__rdiv_fc32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__rdiv_fc32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__rdiv_fc64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__rdiv_fc64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__eq_bool)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__eq_bool)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__gt_bool)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__gt_bool)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__lt_bool)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__lt_bool)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__ge_bool)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__ge_bool)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__le_bool)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__le_bool)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__lor_bool)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__lor_bool)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__lor_int8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__lor_int8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__lor_int16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__lor_int16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__lor_int32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__lor_int32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__lor_int64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__lor_int64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__lor_uint8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__lor_uint8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__lor_uint16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__lor_uint16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__lor_uint32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__lor_uint32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__lor_uint64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__lor_uint64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__lor_fp32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__lor_fp32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__lor_fp64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__lor_fp64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__land_bool)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__land_bool)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__land_int8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__land_int8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__land_int16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__land_int16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__land_int32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__land_int32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__land_int64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__land_int64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__land_uint8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__land_uint8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__land_uint16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__land_uint16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__land_uint32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__land_uint32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__land_uint64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__land_uint64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__land_fp32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__land_fp32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__land_fp64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__land_fp64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__lxor_bool)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__lxor_bool)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__lxor_int8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__lxor_int8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__lxor_int16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__lxor_int16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__lxor_int32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__lxor_int32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__lxor_int64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__lxor_int64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__lxor_uint8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__lxor_uint8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__lxor_uint16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__lxor_uint16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__lxor_uint32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__lxor_uint32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__lxor_uint64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__lxor_uint64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__lxor_fp32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__lxor_fp32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__lxor_fp64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__lxor_fp64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__atan2_fp32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__atan2_fp32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__atan2_fp64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__atan2_fp64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__hypot_fp32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__hypot_fp32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__hypot_fp64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__hypot_fp64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__fmod_fp32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__fmod_fp32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__fmod_fp64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__fmod_fp64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__remainder_fp32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__remainder_fp32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__remainder_fp64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__remainder_fp64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__copysign_fp32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__copysign_fp32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__copysign_fp64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__copysign_fp64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__ldexp_fp32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__ldexp_fp32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__ldexp_fp64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__ldexp_fp64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bor_int8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bor_int8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bor_int16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bor_int16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bor_int32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bor_int32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bor_int64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bor_int64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bor_uint8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bor_uint8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bor_uint16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bor_uint16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bor_uint32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bor_uint32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bor_uint64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bor_uint64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__band_int8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__band_int8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__band_int16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__band_int16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__band_int32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__band_int32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__band_int64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__band_int64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__band_uint8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__band_uint8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__band_uint16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__band_uint16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__band_uint32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__band_uint32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__band_uint64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__band_uint64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bxor_int8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bxor_int8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bxor_int16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bxor_int16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bxor_int32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bxor_int32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bxor_int64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bxor_int64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bxor_uint8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bxor_uint8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bxor_uint16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bxor_uint16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bxor_uint32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bxor_uint32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bxor_uint64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bxor_uint64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bxnor_int8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bxnor_int8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bxnor_int16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bxnor_int16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bxnor_int32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bxnor_int32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bxnor_int64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bxnor_int64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bxnor_uint8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bxnor_uint8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bxnor_uint16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bxnor_uint16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bxnor_uint32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bxnor_uint32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bxnor_uint64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bxnor_uint64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bget_int8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bget_int8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bget_int16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bget_int16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bget_int32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bget_int32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bget_int64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bget_int64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bget_uint8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bget_uint8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bget_uint16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bget_uint16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bget_uint32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bget_uint32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bget_uint64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bget_uint64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bset_int8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bset_int8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bset_int16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bset_int16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bset_int32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bset_int32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bset_int64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bset_int64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bset_uint8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bset_uint8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bset_uint16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bset_uint16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bset_uint32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bset_uint32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bset_uint64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bset_uint64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bclr_int8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bclr_int8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bclr_int16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bclr_int16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bclr_int32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bclr_int32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bclr_int64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bclr_int64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bclr_uint8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bclr_uint8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bclr_uint16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bclr_uint16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bclr_uint32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bclr_uint32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bclr_uint64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bclr_uint64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bshift_int8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bshift_int8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bshift_int16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bshift_int16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bshift_int32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bshift_int32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bshift_int64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bshift_int64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bshift_uint8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bshift_uint8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bshift_uint16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bshift_uint16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bshift_uint32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bshift_uint32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__bshift_uint64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__bshift_uint64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__pow_int8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__pow_int8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__pow_int16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__pow_int16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__pow_int32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__pow_int32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__pow_int64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__pow_int64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__pow_uint8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__pow_uint8)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__pow_uint16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__pow_uint16)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__pow_uint32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__pow_uint32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__pow_uint64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__pow_uint64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__pow_fp32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__pow_fp32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__pow_fp64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__pow_fp64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__pow_fc32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__pow_fc32)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

GrB_Info GB (_subassign_23__pow_fc64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_22__pow_fc64)
(
    GrB_Matrix C,
    const GB_void *ywork_handle
) ;

