//------------------------------------------------------------------------------
// GB_Descriptor_get.h: definitions for GB_Descriptor_get
//------------------------------------------------------------------------------

// SuiteSparse:GraphBLAS, Timothy A. Davis, (c) 2017-2025, All Rights Reserved.
// SPDX-License-Identifier: Apache-2.0

//------------------------------------------------------------------------------

#ifndef GB_DESCRIPTOR_GET_H
#define GB_DESCRIPTOR_GET_H

GrB_Info GB_Descriptor_get      // get the contents of a descriptor
(
    const GrB_Descriptor desc,  // descriptor to query, may be NULL
    bool *C_replace,            // if true replace C before C<M>=Z
    bool *Mask_comp,            // if true use logical negation of M
    bool *Mask_struct,          // if true use the structure of M
    bool *In0_transpose,        // if true transpose first input
    bool *In1_transpose,        // if true transpose second input
    int *AxB_method,            // method for C=A*B
    int *do_sort                // if nonzero, sort in GrB_mxm
) ;

#endif

