//------------------------------------------------------------------------------
// GB_as__include.h: definitions for GB_as__*.c
//------------------------------------------------------------------------------

// SuiteSparse:GraphBLAS, Timothy A. Davis, (c) 2017-2025, All Rights Reserved.
// SPDX-License-Identifier: Apache-2.0

// This file has been automatically generated from Generator/GB_as.h
#include "math/GB_math.h"

GrB_Info GB (_subassign_05d__bool)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const bool Mask_struct,
    const GB_void *scalar,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_06d__bool)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    const bool Mask_struct,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_25__bool)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_05d__int8)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const bool Mask_struct,
    const GB_void *scalar,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_06d__int8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    const bool Mask_struct,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_25__int8)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_05d__int16)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const bool Mask_struct,
    const GB_void *scalar,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_06d__int16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    const bool Mask_struct,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_25__int16)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_05d__int32)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const bool Mask_struct,
    const GB_void *scalar,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_06d__int32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    const bool Mask_struct,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_25__int32)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_05d__int64)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const bool Mask_struct,
    const GB_void *scalar,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_06d__int64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    const bool Mask_struct,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_25__int64)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_05d__uint8)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const bool Mask_struct,
    const GB_void *scalar,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_06d__uint8)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    const bool Mask_struct,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_25__uint8)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_05d__uint16)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const bool Mask_struct,
    const GB_void *scalar,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_06d__uint16)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    const bool Mask_struct,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_25__uint16)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_05d__uint32)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const bool Mask_struct,
    const GB_void *scalar,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_06d__uint32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    const bool Mask_struct,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_25__uint32)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_05d__uint64)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const bool Mask_struct,
    const GB_void *scalar,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_06d__uint64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    const bool Mask_struct,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_25__uint64)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_05d__fp32)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const bool Mask_struct,
    const GB_void *scalar,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_06d__fp32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    const bool Mask_struct,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_25__fp32)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_05d__fp64)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const bool Mask_struct,
    const GB_void *scalar,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_06d__fp64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    const bool Mask_struct,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_25__fp64)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_05d__fc32)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const bool Mask_struct,
    const GB_void *scalar,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_06d__fc32)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    const bool Mask_struct,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_25__fc32)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_05d__fc64)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const bool Mask_struct,
    const GB_void *scalar,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_06d__fc64)
(
    GrB_Matrix C,
    const GrB_Matrix A,
    const bool Mask_struct,
    GB_Werk Werk
) ;

GrB_Info GB (_subassign_25__fc64)
(
    GrB_Matrix C,
    const GrB_Matrix M,
    const GrB_Matrix A,
    GB_Werk Werk
) ;

